/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing authors: Paul Crozier (SNL)
                         Alexander Stukowski
                        + Davide Michieletto

 DMI - this module simulates "infection" of atom types.
 To write this I have started from fix_atom_swap.cpp
 Set variable SPEAK=1 for debugging.
------------------------------------------------------------------ */


#include "fix_atom_recolour.h"
#include <mpi.h>
#include <cmath>
#include <cctype>
#include <cfloat>
#include <cstring>
#include "atom.h"
#include "update.h"
#include "modify.h"
#include "fix.h"
#include "comm.h"
#include "compute.h"
#include "group.h"
#include "domain.h"
#include "region.h"
#include "random_park.h"
#include "random_mars.h"
#include "force.h"
#include "pair.h"
#include "bond.h"
#include "angle.h"
#include "dihedral.h"
#include "improper.h"
#include "kspace.h"
#include "memory.h"
#include "error.h"
#include "neighbor.h"
#include "neigh_request.h"
#include "neigh_list.h"
#include "output.h"
#include <iostream>
#include "citeme.h"

using namespace std;
using namespace LAMMPS_NS;
using namespace FixConst;

static const char cite_fix_recol[] =
"Out-of-equilibrium recolouring by D Michieletto\n\n"
  "@Article{XX,\n"
  " author = {XX},\n"
  " title = {xx},\n"
  " journal = {xx},\n"
  " year =    xx,\n"
  " volume =  xx,\n"
  " pages =   {xx}\n"
  "}\n\n";


/* ---------------------------------------------------------------------- */

FixRecol::FixRecol(LAMMPS *lmp, int narg, char **arg) :
  Fix(lmp, narg, arg),
  idregion(NULL), type_list(NULL), mu(NULL), qtype(NULL),
  sqrt_mass_ratio(NULL), local_swap_iatom_list(NULL),
  local_swap_jatom_list(NULL), local_swap_atom_list(NULL),
  random_equal(NULL), random_unequal(NULL), random(NULL), c_pe(NULL)
{

  if (lmp->citeme) lmp->citeme->add(cite_fix_recol);

  if (narg < 10) error->all(FLERR,"Illegal fix atom/swap command");

  dynamic_group_allow = 1;

  vector_flag = 1;
  size_vector = 2;
  global_freq = 1;
  extvector = 0;
  restart_global = 1;
  time_depend = 1;

//************************************************
// required args
//*************************************************
  nevery = force->inumeric(FLERR,arg[3]);
  ncycles = force->inumeric(FLERR,arg[4]);
  seed = force->inumeric(FLERR,arg[5]);
  alpha = force->numeric(FLERR,arg[6]);
  rmax = force->numeric(FLERR,arg[7]);
//*************************************************
	
  //beta = 1.0/(force->boltz*temperature);

  if (nevery <= 0) error->all(FLERR,"Illegal fix atom/swap command");
  if (ncycles < 0) error->all(FLERR,"Illegal fix atom/swap command");
  if (seed <= 0) error->all(FLERR,"Illegal fix atom/swap command");
  if (alpha < 0) error->all(FLERR,"Illegal fix atom/swap command");
  if (rmax <= 0) error->all(FLERR,"Illegal fix atom/swap command");

  memory->create(type_list,atom->ntypes,"atom/swap:type_list");
  memory->create(mu,atom->ntypes+1,"atom/swap:mu");
  for (int i = 1; i <= atom->ntypes; i++) mu[i] = 0.0;

  // read options from end of input line

  options(narg-8,&arg[8]);

  // random number generator, same for all procs

  random_equal = new RanPark(lmp,seed);

  // random number generator, not the same for all procs

  random_unequal = new RanPark(lmp,seed);

  // initialize Marsaglia RNG with processor-unique seed

  random = new RanMars(lmp,seed + me);
    
  // set up reneighboring

  force_reneighbor = 1;
  next_reneighbor = update->ntimestep + 1;

  //debug
  speak = 1;
  
  // zero out counters

  nswap_attempts = 0.0;
  nswap_successes = 0.0;

  atom_swap_nmax = 0;
  local_swap_atom_list = NULL;
  local_swap_iatom_list = NULL;
  local_swap_jatom_list = NULL;

  // set comm size needed by this Fix

  if (atom->q_flag) comm_forward = 2;
  else comm_forward = 1;

}

/* ----------------------------------------------------------------------
   parse optional parameters at end of input line
------------------------------------------------------------------------- */

void FixRecol::options(int narg, char **arg)
{
  if (narg < 0) error->all(FLERR,"Illegal fix atom/swap command");

  regionflag = 0;
  conserve_ke_flag = 1;
  semi_grand_flag = 0;
  nswaptypes = 0;
  nmutypes = 0;
  iregion = -1;

  int iarg = 0;
  while (iarg < narg) {
    if (strcmp(arg[iarg],"region") == 0) {
      if (iarg+2 > narg) error->all(FLERR,"Illegal fix atom/swap command");
      iregion = domain->find_region(arg[iarg+1]);
      if (iregion == -1)
        error->all(FLERR,"Region ID for fix atom/swap does not exist");
      int n = strlen(arg[iarg+1]) + 1;
      idregion = new char[n];
      strcpy(idregion,arg[iarg+1]);
      regionflag = 1;
      iarg += 2;
    } else if (strcmp(arg[iarg],"ke") == 0) {
      if (iarg+2 > narg) error->all(FLERR,"Illegal fix atom/swap command");
      if (strcmp(arg[iarg+1],"no") == 0) conserve_ke_flag = 0;
      else if (strcmp(arg[iarg+1],"yes") == 0) conserve_ke_flag = 1;
      else error->all(FLERR,"Illegal fix atom/swap command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"semi-grand") == 0) {
      if (iarg+2 > narg) error->all(FLERR,"Illegal fix atom/swap command");
      if (strcmp(arg[iarg+1],"no") == 0) semi_grand_flag = 0;
      else if (strcmp(arg[iarg+1],"yes") == 0) semi_grand_flag = 1;
      else error->all(FLERR,"Illegal fix atom/swap command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"types") == 0) {
      if (iarg+3 > narg) error->all(FLERR,"Illegal fix atom/swap command");
      iarg++;
      while (iarg < narg) {
        if (isalpha(arg[iarg][0])) break;
        if (nswaptypes >= atom->ntypes) error->all(FLERR,"Illegal fix atom/swap command");
        type_list[nswaptypes] = force->numeric(FLERR,arg[iarg]);
        nswaptypes++;
        iarg++;
      }
    } else if (strcmp(arg[iarg],"mu") == 0) {
      if (iarg+2 > narg) error->all(FLERR,"Illegal fix atom/swap command");
      iarg++;
      while (iarg < narg) {
        if (isalpha(arg[iarg][0])) break;
        nmutypes++;
        if (nmutypes > atom->ntypes) error->all(FLERR,"Illegal fix atom/swap command");
        mu[nmutypes] = force->numeric(FLERR,arg[iarg]);
        iarg++;
      }
    } else error->all(FLERR,"Illegal fix atom/swap command");
  }
}

/* ---------------------------------------------------------------------- */

FixRecol::~FixRecol()
{
  memory->destroy(type_list);
  memory->destroy(mu);
  memory->destroy(qtype);
  memory->destroy(sqrt_mass_ratio);
  if (regionflag) delete [] idregion;
  delete random_equal;
  delete random_unequal;
  delete random;
}

/* ---------------------------------------------------------------------- */

int FixRecol::setmask()
{
  int mask = 0;
  mask |= PRE_EXCHANGE;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixRecol::init()
{
  char *id_pe = (char *) "thermo_pe";
  int ipe = modify->find_compute(id_pe);
  c_pe = modify->compute[ipe];

  int *type = atom->type;

  if (nswaptypes < 2)
    error->all(FLERR,"Must specify at least 2 types in fix atom/swap command");

  if (semi_grand_flag) {
    if (nswaptypes != nmutypes)
      error->all(FLERR,"Need nswaptypes mu values in fix atom/swap command");
  } else {
    if (nswaptypes != 2)
      error->all(FLERR,"Only 2 types allowed when not using semi-grand in fix atom/swap command");
    if (nmutypes != 0)
      error->all(FLERR,"Mu not allowed when not using semi-grand in fix atom/swap command");
  }

  for (int iswaptype = 0; iswaptype < nswaptypes; iswaptype++)
    if (type_list[iswaptype] <= 0 || type_list[iswaptype] > atom->ntypes)
      error->all(FLERR,"Invalid atom type in fix atom/swap command");

  // this is only required for non-semi-grand
  // in which case, nswaptypes = 2

  if (atom->q_flag && !semi_grand_flag) {
    double qmax,qmin;
    int firstall,first;
    memory->create(qtype,nswaptypes,"atom/swap:qtype");
    for (int iswaptype = 0; iswaptype < nswaptypes; iswaptype++) {
      first = 1;
      for (int i = 0; i < atom->nlocal; i++) {
        if (atom->mask[i] & groupbit) {
          if (type[i] == type_list[iswaptype]) {
            if (first) {
              qtype[iswaptype] = atom->q[i];
              first = 0;
            } else if (qtype[iswaptype] != atom->q[i])
              error->one(FLERR,"All atoms of a swapped type must have the same charge.");
          }
        }
      }
      MPI_Allreduce(&first,&firstall,1,MPI_INT,MPI_MIN,world);
      if (firstall) error->all(FLERR,"At least one atom of each swapped type must be present to define charges.");
      if (first) qtype[iswaptype] = -DBL_MAX;
      MPI_Allreduce(&qtype[iswaptype],&qmax,1,MPI_DOUBLE,MPI_MAX,world);
      if (first) qtype[iswaptype] = DBL_MAX;
      MPI_Allreduce(&qtype[iswaptype],&qmin,1,MPI_DOUBLE,MPI_MIN,world);
      if (qmax != qmin) error->all(FLERR,"All atoms of a swapped type must have same charge.");
    }
  }

  memory->create(sqrt_mass_ratio,atom->ntypes+1,atom->ntypes+1,"atom/swap:sqrt_mass_ratio");
  for (int itype = 1; itype <= atom->ntypes; itype++)
    for (int jtype = 1; jtype <= atom->ntypes; jtype++)
      sqrt_mass_ratio[itype][jtype] = sqrt(atom->mass[itype]/atom->mass[jtype]);

  // check to see if itype and jtype cutoffs are the same
  // if not, reneighboring will be needed between swaps

  double **cutsq = force->pair->cutsq;
  unequal_cutoffs = false;
  for (int iswaptype = 0; iswaptype < nswaptypes; iswaptype++)
    for (int jswaptype = 0; jswaptype < nswaptypes; jswaptype++)
      for (int ktype = 1; ktype <= atom->ntypes; ktype++)
        if (cutsq[type_list[iswaptype]][ktype] != cutsq[type_list[jswaptype]][ktype])
          unequal_cutoffs = true;

  // check that no swappable atoms are in atom->firstgroup
  // swapping such an atom might not leave firstgroup atoms first

  if (atom->firstgroup >= 0) {
    int *mask = atom->mask;
    int firstgroupbit = group->bitmask[atom->firstgroup];

    int flag = 0;
    for (int i = 0; i < atom->nlocal; i++)
      if ((mask[i] == groupbit) && (mask[i] && firstgroupbit)) flag = 1;

    int flagall;
    MPI_Allreduce(&flag,&flagall,1,MPI_INT,MPI_SUM,world);

    if (flagall)
      error->all(FLERR,"Cannot do atom/swap on atoms in atom_modify first group");
  }

    // need a half neighbor list, built every Nevery steps
    //dmi -- request neighbour list **************************** //
    int irequest = neighbor->request(this,instance_me);
    neighbor->requests[irequest]->pair = 0;
    neighbor->requests[irequest]->fix = 1;
    neighbor->requests[irequest]->occasional = 1;
    // ************************************************************* //
    
    lastcheck = -1; //(? not sure about this if needed)
}

/* ----------------------------------------------------------------------
   attempt Monte Carlo swaps
------------------------------------------------------------------------- */

void FixRecol::pre_exchange()
{
  // just return if should not be called on this timestep

  if (next_reneighbor != update->ntimestep) return;

  energy_stored = energy_full();

  //***********************************
  // HERE THE LIST IS BUILT!
  //************************************
	neighbor->build(1);
       
  int nsuccess = 0;
  if (semi_grand_flag) {
	//**************************************** //
    //important! this updates the list of atoms that can be picked - dmi
    update_semi_grand_atoms_list();
    //**************************************** //
    for (int i = 0; i < ncycles; i++) nsuccess += attempt_semi_grand();
  } else {
    update_swap_atoms_list();
    for (int i = 0; i < ncycles; i++) nsuccess += attempt_swap();
  }

  nswap_attempts += ncycles;
  nswap_successes += nsuccess;

  energy_full();
  next_reneighbor = update->ntimestep + nevery;
  if(speak)fprintf(screen, "end-2?!");
  
}

/* ---------------------------------------------------------------------- */

void FixRecol::init_list(int /*id*/, NeighList *ptr)
{
  list = ptr;
}

/* ----------------------------------------------------------------------
Note: atom charges are assumed equal and so are not updated
------------------------------------------------------------------------- */

int FixRecol::attempt_semi_grand()
{
  if (nswap == 0) return 0;
  
  int i,j,k,m,n,ii,jj,inum,jnum,itype,itypeold,jtype,n1,n2,n3,possible;
  double xtmp,ytmp,ztmp,delx,dely,delz,rsq;
  int *ilist,*jlist,*numneigh,**firstneigh;
  tagint *slist;
  
  // check that all procs have needed ghost atoms within ghost cutoff
  // only if neighbor list has changed since last check
  // needs to be <= test b/c neighbor list could have been re-built in
  //   same timestep as last post_integrate() call, but afterwards
  // NOTE: no longer think is needed, due to error tests on atom->map()
  // NOTE: if delete, can also delete lastcheck and check_ghosts()

  //if (lastcheck <= neighbor->lastcall) check_ghosts();

  // acquire updated ghost atom positions
  // necessary b/c are calling this after integrate, but before Verlet comm

  comm->forward_comm();
  // forward comm of bondcount, so ghosts have it

  commflag = 1;
  comm->forward_comm_fix(this,1);

  // create neighbor list - maybe not needed
  //this command is based on the earlier request
  //NeighList *list = neighbor->lists[irequest];
  //neighbor->build_one(list);
    
  int nlocal = atom->nlocal;
  int nall = atom->nlocal + atom->nghost;

//*****************************************************************
  double **x = atom->x; //positions
  tagint *tag = atom->tag; //this recovers the global id of the atom
  int *mask = atom->mask; //this recovers the info on groups
  int *type = atom->type; //this recovers the info on types
//****************************************************************
	
  neighbor->build_one(list,1);
  inum = list->inum;
  ilist = list->ilist;
  numneigh = list->numneigh;
  firstneigh = list->firstneigh;
    
//****************************************************
//THIS IS MY NEW ARRAY WITH GOOD NEIGHBOURS populated below
//dmi
//**************************************************
  int nmax=1000;
  int newlist[nmax];
    
  double energy_before = energy_stored;
  
  //***************************************** //
  //IMPORTANT - dmi
  //this is to pick the first atom with correct group and type - dmi
  //also I dont want to do this for evey atom but only some picked at random
  i = pick_semi_grand_atom();
  // ************************************** //
  if(speak)fprintf(screen,"picked %d type %d \n", i, atom->type[i]);

//***************************************** //
//CHANGE COLOR TO THAT OF A NEIGHBOUR w prob \alpha
//***************************************** //
double ra=random->uniform();//dmi
if(ra<alpha){ //dmi - recruited conversion
    if (i >= 0) { //dmi
    itype=type[i];
    //search in the neighbourhood of the first atom
    xtmp=x[i][0];
    ytmp=x[i][1];
    ztmp=x[i][2];
    
//*********************
//FIND OUT HOW MANY NEIGHBOURS
//******************************
	jnum = numneigh[i]; //number of neihbours
    jlist = firstneigh[i]; //neighbour list (array)
    if(speak)fprintf(screen,"with %d neighbours \n",jnum);
    
//******************************
//LOOP OVER NEIGHBOURS AND SELECT GOOD ONES
//******************************
	int nnok=0; //# neighbours within my cut-off to be built below
    for(int nn=0;nn<jnum;nn++){
      j=jlist[nn];
      jtype=type[j];
      j &= NEIGHMASK; //keeps "topological" neighbours
      //(i.e. removes them if excluded by special bond command, e.g. in FENE)

//*********************************
//could consider only some (e.g. type=5) at this point
//if they have other type continue through the list
//*****************************
      //if(jtype!=5) continue;
        
	  if(jtype==itype)continue;
      if(speak)fprintf(screen,"looking at %d with tag %d type %d \n",j,tag[j],jtype);
      
      //COMPUTE DISTANCE FROM i
      delx = xtmp - x[j][0];
      dely = ytmp - x[j][1];
      delz = ztmp - x[j][2];
      rsq = delx*delx + dely*dely + delz*delz;

      //IF d<Rcutoff PUT j IN NEWLIST
      if(sqrt(rsq)<rmax){newlist[nnok]=j;nnok++;}
      if(speak)fprintf(screen,"looking at %d at dist %f, with tag %d and type %d \n",j,sqrt(rsq),tag[j],jtype);
      }
      if(speak)fprintf(screen,"number of ok atoms = %d \n",nnok);
      
      if(nnok==0)return 0; //if no good neighb carry on to the next i atom
 
//*****************************
//PICK ONE OF THE GOOD ONES AT RANDOM
//*****************************
      int jtemp=int(nnok*random->uniform());
      j=newlist[jtemp];
      jtype=type[j];
      if(speak)fprintf(screen,"now picked %d (tag %d) with type %d \n",j,tag[j],jtype);
      //if(atom->type[j]==5)cin.get();

//*****************************************************
// comm needed at this point to avoid clashes btwn procs
//******************************************************
      // reverse comm of distsq and partner
      // not needed if newton_pair off since I,J pair was seen by both procs
      commflag = 2;
      if (force->newton_pair) comm->reverse_comm_fix(this);
      // each atom now knows its winning partner
      // for prob check, generate random value for each atom with a bond partner
      // forward comm of partner and random value, so ghosts have it
      commflag = 2;
      comm->forward_comm_fix(this,2);
		
//******************************************************
//THIS IS THE RECOLOURING PART THAT DEPENDS ON THE MODEL
// in this case simple infection
//******************************************************
      itypeold=itype;
      jtype=jtype;

	  //new color/type
	  atom->type[i]=jtype;
//******************************************************

	if(speak)fprintf(screen,"changed %d (tag %d) from type %d to type %d \n",i,tag[i],itypeold,atom->type[i]);
          //cin.get();
          update_semi_grand_atoms_list();
    } //closed if i>=0
  } //closed if ra<alpha
//********************************************************
// if rand > alpha ->  noisy conversion
//*****************************************************
  else {
    if (i >= 0) {
    itype=type[i];
    itypeold=type[i];
    
    /*
	 // reverse comm of distsq and partner
    // not needed if newton_pair off since I,J pair was seen by both procs
    commflag = 2;
    if (force->newton_pair) comm->reverse_comm_fix(this);
    // each atom now knows its winning partner
    // for prob check, generate random value for each atom with a bond partner
    // forward comm of partner and random value, so ghosts have it
    commflag = 2;
    comm->forward_comm_fix(this,2);
    */
		
//***********************************************
//equal probability of going to any other color
//************************************************
	double raa=random->uniform();
    atom->type[i]=int(random->uniform()*1000)%4+1;// [1:4]
	
    if(speak)fprintf(screen,"doing random recol of %d (type %d) into %d \n",i,itypeold,itype); //cin.get();
    }
  
  if(speak)fprintf(screen,"noisy conversion ended");
  update_semi_grand_atoms_list();
  }
   
   // communicate final types (?)
   commflag = 3;
   comm->forward_comm_fix(this);

  if (force->kspace) force->kspace->qsum_qsq();
  double energy_after = energy_full();

  int success = 1; //either recruited or noisy
//*************************
  /* IF ENERGY IS NEEDED */
  /*
    if (i >= 0)
      if (random_unequal->uniform() <
        exp(beta*(energy_before - energy_after
              + mu[jtype] - mu[itype]))) success = 1;
 */
//*****************************
  int success_all = 0;
  //this takes the maximum of all "successes"
  MPI_Allreduce(&success,&success_all,1,MPI_INT,MPI_MAX,world);

  if (success_all) {
    energy_stored = energy_after;
    if (conserve_ke_flag) {
      if (i >= 0) {
        atom->v[i][0] *= sqrt_mass_ratio[itype][jtype];
        atom->v[i][1] *= sqrt_mass_ratio[itype][jtype];
        atom->v[i][2] *= sqrt_mass_ratio[itype][jtype];
      }
    }
    return 1;
  } else {
    if (i >= 0) { //dmi to account r > alpha cases
      atom->type[i] = itype;
    }
    if (force->kspace) force->kspace->qsum_qsq();
    energy_stored = energy_before;

    if (unequal_cutoffs) {
      if (domain->triclinic) domain->x2lamda(atom->nlocal);
      domain->pbc();
      comm->exchange();
      comm->borders();
      if (domain->triclinic) domain->lamda2x(atom->nlocal+atom->nghost);
      if (modify->n_pre_neighbor) modify->pre_neighbor();
      neighbor->build(1);
    } else {
      comm->forward_comm_fix(this);
    }
  }
    
  return 0;
}

/* ----------------------------------------------------------------------
------------------------------------------------------------------------- */

int FixRecol::attempt_swap()
{
  return 0;
}

/* ----------------------------------------------------------------------
   compute system potential energy
------------------------------------------------------------------------- */

double FixRecol::energy_full()
{
  double total_energy=0;
  
  int eflag = 1;
  int vflag = 0;

  if (modify->n_pre_neighbor) modify->pre_neighbor();
  if (modify->n_pre_force) modify->pre_force(vflag);

  if (force->pair) force->pair->compute(eflag,vflag);

  if (atom->molecular) {
    if (force->bond) force->bond->compute(eflag,vflag);
    if (force->angle) force->angle->compute(eflag,vflag);
    if (force->dihedral) force->dihedral->compute(eflag,vflag);
    if (force->improper) force->improper->compute(eflag,vflag);
  }

  if (force->kspace) force->kspace->compute(eflag,vflag);

  if (modify->n_post_force) modify->post_force(vflag);
  if (modify->n_end_of_step) modify->end_of_step();

  update->eflag_global = update->ntimestep;
  total_energy = c_pe->compute_scalar();
  
  return total_energy;
}

/* ----------------------------------------------------------------------
------------------------------------------------------------------------- */

int FixRecol::pick_semi_grand_atom()
{
  int i = -1;
  int iwhichglobal = static_cast<int> (nswap*random_equal->uniform());
  if ((iwhichglobal >= nswap_before) &&
      (iwhichglobal < nswap_before + nswap_local)) {
    int iwhichlocal = iwhichglobal - nswap_before;
    i = local_swap_atom_list[iwhichlocal];
    if(speak)fprintf(screen,"nswap_before %d and nswap_local %d \n",nswap_before, nswap_local);
    if(speak)fprintf(screen,"considering %d atoms .. picked %d type %d \n",nswap,i,atom->type[i]);
  }

  return i;
}

/* ----------------------------------------------------------------------
------------------------------------------------------------------------- */

int FixRecol::pick_i_swap_atom()
{
  int i = -1;
  int iwhichglobal = static_cast<int> (niswap*random_equal->uniform());
  if ((iwhichglobal >= niswap_before) &&
      (iwhichglobal < niswap_before + niswap_local)) {
    int iwhichlocal = iwhichglobal - niswap_before;
    i = local_swap_iatom_list[iwhichlocal];
  }

  return i;
}

/* ----------------------------------------------------------------------
------------------------------------------------------------------------- */

int FixRecol::pick_j_swap_atom()
{
  int j = -1;
  int jwhichglobal = static_cast<int> (njswap*random_equal->uniform());
  if ((jwhichglobal >= njswap_before) &&
      (jwhichglobal < njswap_before + njswap_local)) {
    int jwhichlocal = jwhichglobal - njswap_before;
    j = local_swap_jatom_list[jwhichlocal];
  }

  return j;
}

/* ----------------------------------------------------------------------
   update the list of gas atoms
------------------------------------------------------------------------- */

void FixRecol::update_semi_grand_atoms_list()
{
  int nlocal = atom->nlocal; //get number of local atoms
  double **x = atom->x; //get positions

  if (atom->nmax > atom_swap_nmax) {
    memory->sfree(local_swap_atom_list);
    atom_swap_nmax = atom->nmax;
    local_swap_atom_list = (int *) memory->smalloc(atom_swap_nmax*sizeof(int),
     "MCSWAP:local_swap_atom_list");
  }

  nswap_local = 0;

  //if there is a region constraints
  if (regionflag) {

    for (int i = 0; i < nlocal; i++) {
      if (domain->regions[iregion]->match(x[i][0],x[i][1],x[i][2]) == 1) {
        if (atom->mask[i] & groupbit) {
          int iitype = atom->type[i];
          int iswaptype;
          for (iswaptype = 0; iswaptype < nswaptypes; iswaptype++)
            if (iitype == type_list[iswaptype]) break;
          if (iswaptype == nswaptypes) continue;
          local_swap_atom_list[nswap_local] = i;
          nswap_local++;
        }
      }
    }

  } else { //if there is no region constraint
    for (int i = 0; i < nlocal; i++) { //loops over local atoms
      if (atom->mask[i] & groupbit) { //select atoms only in the group specified (use all)
          int iitype = atom->type[i];
          int iswaptype;
          for (iswaptype = 0; iswaptype < nswaptypes; iswaptype++) //nswaptypes is the #
                                                                 //of types that can be swapped
           if (iitype == type_list[iswaptype]) break;
           if (iswaptype == nswaptypes) continue;
          
          //IMPORTANT dmi ******************************* //
          //can hard cod the constraint that the atoms picked need to have certain type
          //if this is e.g. type 5, then continue through the local atom list
          //this limits the number of atoms from which it is possible to pick
          //if(iitype == 5) continue;
          //************************************************ //
          
          //create list of atoms that can be swapped
          local_swap_atom_list[nswap_local] = i;
          nswap_local++;
      }
    }
  }

  MPI_Allreduce(&nswap_local,&nswap,1,MPI_INT,MPI_SUM,world);
  MPI_Scan(&nswap_local,&nswap_before,1,MPI_INT,MPI_SUM,world);
  nswap_before -= nswap_local;
}


/* ----------------------------------------------------------------------
   update the list of gas atoms
------------------------------------------------------------------------- */

void FixRecol::update_swap_atoms_list()
{
  int nlocal = atom->nlocal;
  int *type = atom->type;
  double **x = atom->x;

  if (atom->nmax > atom_swap_nmax) {
    memory->sfree(local_swap_iatom_list);
    memory->sfree(local_swap_jatom_list);
    atom_swap_nmax = atom->nmax;
    local_swap_iatom_list = (int *) memory->smalloc(atom_swap_nmax*sizeof(int),
     "MCSWAP:local_swap_iatom_list");
    local_swap_jatom_list = (int *) memory->smalloc(atom_swap_nmax*sizeof(int),
     "MCSWAP:local_swap_jatom_list");
  }

  niswap_local = 0;
  njswap_local = 0;

  //if region constraint
  if (regionflag) {

    for (int i = 0; i < nlocal; i++) {
      if (domain->regions[iregion]->match(x[i][0],x[i][1],x[i][2]) == 1) {
        if (atom->mask[i] & groupbit) {
          if (type[i] ==  type_list[0]) {
            local_swap_iatom_list[niswap_local] = i;
            niswap_local++;
          } else if (type[i] ==  type_list[1]) {
            local_swap_jatom_list[njswap_local] = i;
            njswap_local++;
          }
        }
      }
    }

  } else { //if region is not a constraint
    for (int i = 0; i < nlocal; i++) {
      if (atom->mask[i] & groupbit) {
        if (type[i] ==  type_list[0]) {
          local_swap_iatom_list[niswap_local] = i;
          niswap_local++;
        } else if (type[i] ==  type_list[1]) {
          local_swap_jatom_list[njswap_local] = i;
          njswap_local++;
        }
      }
    }
  }

  MPI_Allreduce(&niswap_local,&niswap,1,MPI_INT,MPI_SUM,world);
  MPI_Scan(&niswap_local,&niswap_before,1,MPI_INT,MPI_SUM,world);
  niswap_before -= niswap_local;

  MPI_Allreduce(&njswap_local,&njswap,1,MPI_INT,MPI_SUM,world);
  MPI_Scan(&njswap_local,&njswap_before,1,MPI_INT,MPI_SUM,world);
  njswap_before -= njswap_local;
}

/* ---------------------------------------------------------------------- */

int FixRecol::pack_forward_comm(int n, int *list, double *buf, int /*pbc_flag*/, int * /*pbc*/)
{
  int i,j,m;

  int *type = atom->type;
  double *q = atom->q;

  m = 0;

  if (atom->q_flag) {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = type[j];
      buf[m++] = q[j];
    }
  } else {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = type[j];
    }
  }

  return m;
}

/* ---------------------------------------------------------------------- */

void FixRecol::unpack_forward_comm(int n, int first, double *buf)
{
  int i,m,last;

  int *type = atom->type;
  double *q = atom->q;

  m = 0;
  last = first + n;

  if (atom->q_flag) {
    for (i = first; i < last; i++) {
      type[i] = static_cast<int> (buf[m++]);
      q[i] = buf[m++];
    }
  } else {
    for (i = first; i < last; i++)
      type[i] = static_cast<int> (buf[m++]);
  }
}

/* ----------------------------------------------------------------------
  return acceptance ratio
------------------------------------------------------------------------- */

double FixRecol::compute_vector(int n)
{
  if (n == 0) return nswap_attempts;
  if (n == 1) return nswap_successes;
  return 0.0;
}

/* ----------------------------------------------------------------------
   memory usage of local atom-based arrays
------------------------------------------------------------------------- */

double FixRecol::memory_usage()
{
  double bytes = atom_swap_nmax * sizeof(int);
  return bytes;
}

/* ----------------------------------------------------------------------
   pack entire state of Fix into one write
------------------------------------------------------------------------- */

void FixRecol::write_restart(FILE *fp)
{
  int n = 0;
  double list[6];
  list[n++] = random_equal->state();
  list[n++] = random_unequal->state();
  list[n++] = ubuf(next_reneighbor).d;
  list[n++] = nswap_attempts;
  list[n++] = nswap_successes;
  list[n++] = ubuf(update->ntimestep).d;

  if (comm->me == 0) {
    int size = n * sizeof(double);
    fwrite(&size,sizeof(int),1,fp);
    fwrite(list,sizeof(double),n,fp);
  }
}

/* ----------------------------------------------------------------------
   use state info from restart file to restart the Fix
------------------------------------------------------------------------- */

void FixRecol::restart(char *buf)
{
  int n = 0;
  double *list = (double *) buf;

  seed = static_cast<int> (list[n++]);
  random_equal->reset(seed);

  seed = static_cast<int> (list[n++]);
  random_unequal->reset(seed);

  next_reneighbor = (bigint) ubuf(list[n++]).i;

  nswap_attempts = static_cast<int>(list[n++]);
  nswap_successes = static_cast<int>(list[n++]);

  bigint ntimestep_restart = (bigint) ubuf(list[n++]).i;
  if (ntimestep_restart != update->ntimestep)
    error->all(FLERR,"Must not reset timestep when restarting fix atom/swap");
}

